"""Torrent handling."""

from pathlib import Path
from typing import List
from urllib.request import urlopen

from torrentool.api import Torrent


def create_torrent_file(path: Path, webseeds: List[str], output: Path) -> str:
    """Create a torrent file for a path and write it to an output directory.

    Parameters
    ----------
    path: Path
        The path for which to create a torrent file
    webseeds: List[str]
        The list of webseeds to add to the torrent file
    output: Path
        A path to write the .torrent file to

    Returns
    -------
    str
        A string representing the name of the torrent file
    """
    torrent = Torrent.create_from(path)
    torrent.webseeds = webseeds
    torrent.to_file(output)

    return output.name


def get_webseeds(
    artifact_type: str,
    mirrorlist_url: str,
    version: str,
) -> List[str]:
    """Assemble webseeds for an artifact type by querying a mirrorlist URL.

    Parameters
    ----------
    artifact_type: str
        The artifact type to create the webseeds for
    mirrorlist_url: str
        A URL used to retrieve the list of mirrors
    version: str
        The version of the artifact type to create the webseeds for

    Returns
    -------
    List[str]
        A list of strings representing webseeds for an artifact type in a specific version
    """
    webseeds: List[str] = []
    url = urlopen(mirrorlist_url)
    for line in url.read().decode("utf-8").split("\n"):
        if "#Server = " in line:
            webseeds += [
                line.replace(
                    "#Server = ",
                    "",
                ).replace(
                    "$repo/os/$arch",
                    f"releases/{artifact_type}/{version}/",
                )
            ]

    return webseeds



"""Release handling."""

from typing import List, Optional

from pydantic import BaseModel


class Metric(BaseModel):
    """A pydantic model describing a metric.

    Attributes
    ----------
    name: str
        A name for the metric
    description: str
        A description for the metric
    """

    name: str
    description: str


class SizeMetric(Metric):
    """A pydantic model describing a size metric.

    Attributes
    ----------
    size: float
    """

    size: float


class AmountMetric(Metric):
    """A pydantic model describing an amount metric.

    Attributes
    ----------
    amount: int
    """

    amount: int


class VersionMetric(Metric):
    """A pydantic model describing a version metric.

    Attributes
    ----------
    version: str
    """

    version: str


class Release(BaseModel):
    """A pydantic model describing a release.

    Attributes
    ----------
    name: str
        The name of the artifact type
    version: str
        The version of the artifact
    files: List[str]
        A list of files that belong to the release
    amount_metrics: List[AmountMetric]
        A list of AmountMetric instances that are related to the release
    size_metrics: List[SizeMetric]
        A list of SizeMetric instances that are related to the release
    version_metrics: List[VersionMetric]
        A list of VersionMetric instances that are related to the release
    torrent_file: str
        A string representing the name of a torrent file for the release
    developer: str
        The name and mail address of the developer creating the release
    pgp_public_key: str
        The long format of the PGP public key ID used to sign artifacts in the release
    """

    name: str
    version: str
    files: List[str]
    amount_metrics: List[AmountMetric]
    size_metrics: List[SizeMetric]
    version_metrics: List[VersionMetric]
    torrent_file: Optional[str] = None
    developer: str
    pgp_public_key: str



"""Configuration handling."""
from __future__ import annotations

from collections import Counter
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type

try:
    import tomllib
except ModuleNotFoundError:  # pragma: no cover
    import tomli as tomllib  # type: ignore[no-redef]

from dotenv import dotenv_values
from email_validator import EmailNotValidError, validate_email
from pydantic import BaseModel, field_validator, model_validator
from pydantic.fields import FieldInfo
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)
from xdg.BaseDirectory import xdg_config_home

MAKEPKG_CONFIGS = [
    Path("/etc/makepkg.conf"),
    Path(f"{xdg_config_home}/pacman/makepkg.conf"),
    Path("~/.makepkg.conf"),
]

PROJECTS_CONFIGS = [
    Path("/etc/arch-release-promotion/projects.toml"),
    Path(f"{xdg_config_home}/arch-release-promotion/projects.toml"),
]

PROJECTS_SYNC_DIR = Path("/var/lib/arch-release-sync/")
PROJECTS_SYNC_BACKLOG = 3


class ReleaseConfig(BaseModel):
    """A pydantic model describing the configuration of a project's release.

    Attributes
    ----------
    name: str
        The name of the release (type)
    version_metrics: Optional[List[str]]
        A list of names that identify labels in metric samples of type "info", that should be extracted from the
        project's metrics file
    size_metrics: Optional[List[str]]
        A list of names that identify labels in metric samples of type "gauge", that should be extracted from the
        project's metrics file
    amount_metrics: Optional[List[str]]
        A list of names that identify labels in metric samples of type "summary", that should be extracted from the
        project's metrics file
    extensions_to_sign: List[str]
        A list of file extensions for which to create detached signatures
    create_torrent: bool
        A bool indicating whether to create a torrent file for the release (defaults to False)
    """

    name: str
    version_metrics: Optional[List[str]] = None
    size_metrics: Optional[List[str]] = None
    amount_metrics: Optional[List[str]] = None
    extensions_to_sign: List[str]
    create_torrent: bool = False


class SyncConfig(BaseModel):
    """A pydantic model describing configuration for synchronization.

    Attributes
    ----------
    backlog: int
        The backlog of releases to retain at a maximum (defaults to PROJECTS_SYNC_BACKLOG when a SysConfig instance is
        used in a Projects instance or a ProjectConfig instance).
    directory: Path
        A directory into which to sync project release types and their respective releases (defaults to
        PROJECTS_SYNC_DIR when a SysConfig instance is used in a Projects instance or a ProjectConfig instance).
    last_updated_file: Optional[Path]
        The optional path to a file, that is used to write a timestamp to, if the synchronization of a project leads to
        the changing of data on disk (defaults to None).
    temp_in_sync_dir: bool
        A bool specifying whether to download temporary data to temporary directories below sync_dir (defaults to True).
        If False is specified the temporary data is downloaded to the respective user's temporary directory.
    """

    backlog: Optional[int] = None
    directory: Optional[Path] = None
    last_updated_file: Optional[Path] = None
    temp_in_sync_dir: bool = True


class ProjectConfig(BaseModel):
    """A pydantic model describing the configuration of a project.

    Attributes
    ----------
    name: str
        The name of the project
    job_name: str
        The project's job, that offers release artifacts
    output_dir: Path
        The project's configured output directory for release artifacts
    metrics_file: Path
        The project's metrics file
    releases: List[ReleaseConfig]
        The project's list of releases
    sync_config: Optional[SyncConfig]
        An optional SyncConfig instance, which is used to override any global defaults.
    """

    name: str
    job_name: str
    output_dir: Path
    metrics_file: Path
    releases: List[ReleaseConfig]
    sync_config: Optional[SyncConfig] = None


class Projects(BaseSettings):
    """A pydantic BaseSettings class to describe sets of project settings.

    Attributes
    ----------
    projects: List[ProjectConfig]
        A list of project configurations
    sync_config: Optional[SyncConfig]
        An optional SyncConfig instance, which is used to override any implicit defaults and sets defaults for all
        ProjectConfig instances in projects.
    """

    projects: List[ProjectConfig]
    sync_config: Optional[SyncConfig] = None

    model_config = SettingsConfigDict(extra="ignore")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """Customize the configuration source assembly."""
        return (TomlConfig(settings_cls),)

    @field_validator("projects")
    @classmethod
    def validate_project_releases_unique(cls, projects: List[ProjectConfig]) -> List[ProjectConfig]:
        """Validate the list of ProjectConfig instances to only contain uniquely named ReleaseConfig instances.

        Parameters
        ----------
        projects: List[ProjectConfig]
            A list of ProjectConfig instances

        Raises
        ------
        ValueError
            If a non-unique name is encountered among any ReleaseConfig instance

        Returns
        -------
        List[ProjectConfig]
            The unmodified list of ProjectConfig instances
        """
        release_types: List[ReleaseConfig] = []
        for project in projects:
            release_types += project.releases
        names = [release_type.name for release_type in release_types]

        if len(set(names)) < len(names):
            duplicates = [name for name, count in Counter(names).items() if count > 1]
            raise ValueError(
                f"The following release type {'name' if len(duplicates) == 1 else 'names'} "
                f"{'is' if len(duplicates) == 1 else 'are'} not unique: "
                f"{duplicates[0] if len(duplicates) == 1 else duplicates}"
            )

        return projects

    @model_validator(mode="before")
    def validate_projects(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate the list of ProjectConfig instances and override defaults.

        If a ProjectConfig does not specify a SysConfig, override it with the global SysConfig. If a global SysConfig
        does not exist, override with an implicit default where directory defaults to PROJECTS_SYNC_DIR and backlog to
        PROJECTS_SYNC_BACKLOG.

        Parameters
        ----------
        values: Dict[str, Any]
            A dict with all values of the Projects instance

        Returns
        -------
        values: Dict[str, Any]
            The (potentially modified) dict with all values of the Projects instance
        """
        default_sync_config = SyncConfig(
            directory=PROJECTS_SYNC_DIR,
            backlog=PROJECTS_SYNC_BACKLOG,
            last_updated_file=None,
        )

        projects: List[Dict[str, Any]] = values.get("projects") or []
        sync_config: Optional[Dict[str, Any]] = values.get("sync_config")

        # merge global sync_config with defaults
        if sync_config:
            global_sync_config = sync_config | default_sync_config.model_dump()
        else:
            global_sync_config = default_sync_config.model_dump()

        for project in projects:
            if not project.get("sync_config"):
                project["sync_config"] = global_sync_config
            else:
                project["sync_config"] = project.get("sync_config") | global_sync_config  # type: ignore[operator]

        return {"projects": projects, "sync_config": sync_config}

    def get_project(self, name: str) -> ProjectConfig:
        """Return a ProjectConfig by name.

        Parameters
        ----------
        name: str
            A string that matches the name attribute of a ProjectConfig

        Raises
        ------
        RuntimeError
            If no ProjectConfig instance of the given name can be found

        Returns
        -------
        ProjectConfig
            The configuration identified by the provided name
        """
        for project in self.projects:
            if project.name == name:
                return project

        raise RuntimeError(f"No project configuration of the name '{name}' can be found!")


class TomlConfig(PydanticBaseSettingsSource):
    """A TOML config reader."""

    def get_field_value(self, field: FieldInfo, field_name: str) -> Any:  # pragma: no cover
        """
        Get the value, the key for model creation, and a flag to determine whether value is complex.

        Parameters
        ----------
            field: The field.
            field_name: The field name.

        Returns
        -------
            A tuple containing the key, value and a boolean to determine whether value is complex.
        """
        pass

    def __call__(self) -> dict[str, Any]:
        """Read all available projects.toml files."""
        output_dict: dict[str, Any] = {}
        config_files: List[Path] = []
        for config_file in PROJECTS_CONFIGS:
            if config_file.exists():
                config_files += [config_file]

        if not config_files:
            raise RuntimeError("There are no project configuration files!")

        for config_file in config_files:
            with open(config_file, "rb") as file:
                file_dict = tomllib.load(file)
                output_dict = output_dict | file_dict

        return output_dict


class MakepkgConfig(PydanticBaseSettingsSource):
    """A makepkg config reader."""

    def get_field_value(self, field: FieldInfo, field_name: str) -> Any:  # pragma: no cover
        """
        Get the value, the key for model creation, and a flag to determine whether value is complex.

        Parameters
        ----------
            field: The field.
            field_name: The field name.

        Returns
        -------
            A tuple containing the key, value and a boolean to determine whether value is complex.
        """
        pass

    def __call__(self) -> dict[str, Any]:
        """Read all available projects.toml files."""
        config: Dict[str, Optional[str]] = {}
        for config_file in MAKEPKG_CONFIGS:
            config.update(dotenv_values(config_file.expanduser()))

        return config


class Settings(BaseSettings):
    """A class to describe configuration.

    Attributes
    ----------
    GITLAB_URL: str
        A URL for a GitLab upstream (defaults to "https://gitlab.archlinux.org")
    GPGKEY: str
        The PGP key id to use for artifact signatures
    MIRRORLIST_URL: str
        A URL to derive a mirrorlist from (defaults to
        "https://archlinux.org/mirrorlist/?country=all&protocol=http&protocol=https")
    PACKAGER: str
        The packager name and mail address (UID) to use for artifact signatures
    PRIVATE_TOKEN: Optional[str]
        An optional private token to use for authenticating against an upstream
    """

    GITLAB_URL: str = "https://gitlab.archlinux.org"
    GPGKEY: str
    MIRRORLIST_URL: str = "https://archlinux.org/mirrorlist/?country=all&protocol=http&protocol=https"
    PACKAGER: str
    PRIVATE_TOKEN: Optional[str] = None

    model_config = SettingsConfigDict(extra="ignore")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """Customize the sources for configuration."""
        return (
            MakepkgConfig(settings_cls),
            env_settings,
        )

    @field_validator("PACKAGER")
    @classmethod
    def validate_packager(cls, packager: str) -> str:
        """Validate the PACKAGER attribute.

        Parameters
        ----------
        packager: str
            The packager string to validate

        Raises
        ------
        ValueError
            If the packager string is not valid

        Returns
        -------
        str
            A valid packager string
        """
        if len(packager) == 0:
            raise ValueError("The PACKAGER string can not be empty.")
        if not ("<" in packager and ">" in packager):
            raise ValueError(f"The PACKAGER string has to define a mail address: {packager}")
        split_packager = packager.replace(">", "").split("<")
        if len(split_packager[0]) < 1:
            raise ValueError(f"The PACKAGER string has to define a name: {packager}")
        try:
            validate_email(split_packager[1])
        except EmailNotValidError as e:
            raise ValueError(f"The PACKAGER string has to define a valid mail address: {packager}\n{e}")

        return packager

    @field_validator("GPGKEY")
    @classmethod
    def validate_gpgkey(cls, gpgkey: str) -> str:
        """Validate the GPGKEY attribute.

        Parameters
        ----------
        gpgkey: str
            The packager string to validate

        Raises
        ------
        ValueError
            If the gpgkey string is not valid

        Returns
        -------
        str
            A gpgkey string in long-format
        """
        if len(gpgkey) < 40:
            raise ValueError(f"The GPGKEY string has to represent a PGP key ID in long format (40 chars): {gpgkey}")

        return gpgkey

    @field_validator("PRIVATE_TOKEN")
    @classmethod
    def validate_private_token(cls, private_token: Optional[str]) -> Optional[str]:
        """Validate the PRIVATE_TOKEN attribute.

        Parameters
        ----------
        private_token: str
            The private token string to validate

        Raises
        ------
        ValueError
            If the private token string is not valid

        Returns
        -------
        str
            A gpgkey string in long-format
        """
        if private_token is None:
            return None

        if len(private_token) < 20:
            raise ValueError("The PRIVATE_TOKEN string has to represent a valid private token (20 chars).")

        return private_token



"""Local file handling."""

import shutil
import tempfile
import time
import zipfile
from pathlib import Path
from typing import List, Optional, Tuple

import orjson
from prometheus_client.parser import text_fd_to_metric_families
from pydantic import BaseModel, ConfigDict

from arch_release_promotion.config import ProjectConfig, Settings
from arch_release_promotion.gitlab import Upstream
from arch_release_promotion.release import (
    AmountMetric,
    Release,
    SizeMetric,
    VersionMetric,
)

TEMP_DIR_PREFIX = "arp-"


def files_in_dir(path: Path) -> List[str]:
    """Return the files in a directory as a list of strings.

    Parameters
    ----------
    path: Path
        The path for which to list files

    Raises
    ------
    RuntimeError
        If the provided path is not a directory

    Returns
    -------
    List[str]
        A list of strings representing the files in a directory
    """
    if not path.is_dir():
        raise RuntimeError(f"The path is not a path: {path}")

    return [child.name for child in path.iterdir()]


def get_version_from_artifact_release_dir(path: Path) -> str:
    """Extrapolate the version of a given release path.

    Parameters
    ----------
    path: Path
        The path to a release

    Raises
    ------
    RuntimeError
        If the provided path is not a directory

    Returns
    -------
    str
        The version string for the release
    """
    if not path.is_dir():
        raise RuntimeError(f"The path for the release is not a directory: {path}")

    version = path.name.split("-")[1]
    if len(version) < 1:
        raise RuntimeError(f"The extracted version string '{version}' from path '{path}' is not valid.")

    return version


def create_temp_dir() -> Path:
    """Create a temporary directory.

    Returns
    -------
    Path
        The path to the directory
    """
    return Path(tempfile.mkdtemp(prefix=TEMP_DIR_PREFIX))


def remove_temp_dir(path: Path) -> None:
    """Remove a temporary directory recursively.

    Parameters
    ----------
    path: Path
        The path to a directory to remove

    Raises
    ------
    RuntimeError
        If the to be removed path is not a directory or if its name does not contain TEMP_DIR_PREFIX
    """
    if not path.is_dir():
        raise RuntimeError(f"The path to remove is not a directory: {path}")

    if TEMP_DIR_PREFIX not in path.name:
        raise RuntimeError(f"Can not remove a temporary directory, that is not created by the same tool: {path}")

    shutil.rmtree(path=path)


def extract_zip_file_to_parent_dir(path: Path) -> None:
    """Extract the contents of a ZIP file to the parent directory of that file.

    Parameters
    ----------
    path: Path
        The path to a ZIP file

    Raises
    ------
    RuntimeError
        If the path does not specify a valid zip file
    """
    if not zipfile.is_zipfile(path):
        raise RuntimeError(f"The file is not a ZIP file: {path}")

    with zipfile.ZipFile(file=path, mode="r") as zip_file:
        zip_file.extractall(path=path.parent)


def copy_signatures(source: Path, destination: Path) -> None:
    """Copy any signature files from a source directory to a destination directory.

    Parameters
    ----------
    source: Path
        The source directory to copy signature files from
    destination: Path
        The destination directory to copy signature files to

    Raises
    ------
    RuntimeError
        If either source or destination is not a directory
    """
    for path in [source, destination]:
        if not path.is_dir():
            raise RuntimeError("The specified path is not a directory: {path}")

    for file in source.iterdir():
        if file.suffix in [".sig"]:
            shutil.copy(src=file, dst=destination)


def write_release_info_to_file(release: Release, path: Path) -> None:
    """Write a Release instance to a JSON file.

    Parameters
    ----------
    release: Release
        A release instance that will be serialized to JSON
    path: Path
        The file to write the JSON string to

    Raises
    ------
    IsADirectoryError:
        If the path is a directory
    OSError:
        If the file path is not writable
    """
    with open(path, "wb") as file:
        file.write(
            orjson.dumps(
                release.model_dump(), option=orjson.OPT_INDENT_2 | orjson.OPT_APPEND_NEWLINE | orjson.OPT_SORT_KEYS
            )
        )


def load_release_from_json_payload(path: Path) -> Release:
    """Read a JSON payload and return it as a Release instance.

    Parameters
    ----------
    path: Path
        The path to a file containing a JSON payload

    Returns
    -------
    Release
        A Release instance reflecting the data from the JSON payload
    """
    with open(path, "r") as file:
        return Release(**orjson.loads(file.read()))


def write_zip_file_to_parent_dir(path: Path, name: str = "promotion", format: str = "zip") -> None:
    """Create ZIP file of all contents in a directory and write it to the directory's parent.

    Parameters
    ----------
    path: Path
        The path to add to the ZIP file
    name: str
        The name to use for the compressed file (defaults to "promotion")
    format: str
        The compressed file format to use (defaults to "zip")
    """
    known_formats = ["zip", "tar", "gztar", "bztar", "xztar"]
    if len(name) < 1:
        raise RuntimeError("The file name has to be at least one char long, but empty string was provided.")
    if format not in known_formats:
        raise RuntimeError(f"The format must be one of {known_formats}, but {format} is provided.")

    shutil.make_archive(base_name=str(path.parent / Path(name)), format=format, root_dir=path)


def read_metrics_file(
    path: Path,
    version_metrics_names: Optional[List[str]],
    size_metrics_names: Optional[List[str]],
    amount_metrics_names: Optional[List[str]],
) -> Tuple[List[AmountMetric], List[SizeMetric], List[VersionMetric]]:
    """Read a metrics file that contains openmetrics based metrics and return those that match the respective keywords.

    Parameters
    ----------
    path: Path
        The path of the file to read
    version_metrics_names: Optional[List[str]]
        A list of metric names to search for in the labels of metric samples of type "info"
    size_metrics_names: Optional[List[str]],
        A list of metric names to search for in the labels of metric samples of type "gauge"
    amount_metrics_names: Optional[List[str]],
        A list of metric names to search for in the labels of metric samples of type "summary"

    Returns
    -------
    Tuple[List[AmountMetric], List[SizeMetric], List[VersionMetric]]:
        A Tuple with lists of AmountMetric, SizeMetric and VersionMetric instances derived from the input file
    """
    amount_metrics: List[AmountMetric] = []
    size_metrics: List[SizeMetric] = []
    version_metrics: List[VersionMetric] = []

    if path.exists():
        with open(path, "r") as file:
            for metric in text_fd_to_metric_families(file):
                for sample in metric.samples:
                    if (
                        version_metrics_names
                        and metric.type == "info"
                        and metric.name == "version_info"
                        and sample.labels.get("name") in version_metrics_names
                        and sample.labels.get("description")
                        and sample.labels.get("version")
                    ):
                        version_metrics += [
                            VersionMetric(
                                name=sample.labels.get("name"),  # type: ignore[arg-type]
                                description=sample.labels.get("description"),  # type: ignore[arg-type]
                                version=sample.labels.get("version"),  # type: ignore[arg-type]
                            )
                        ]
                    if (
                        size_metrics_names
                        and metric.type == "gauge"
                        and metric.name == "artifact_bytes"
                        and sample.labels.get("name") in size_metrics_names
                        and sample.labels.get("description")
                        and sample.value
                    ):
                        size_metrics += [
                            SizeMetric(
                                name=sample.labels.get("name"),  # type: ignore[arg-type]
                                description=sample.labels.get("description"),  # type: ignore[arg-type]
                                size=sample.value,
                            )
                        ]
                    if (
                        amount_metrics_names
                        and metric.type == "summary"
                        and metric.name == "data_count"
                        and sample.labels.get("name") in amount_metrics_names
                        and sample.labels.get("description")
                        and sample.value
                    ):
                        amount_metrics += [
                            AmountMetric(
                                name=sample.labels.get("name"),  # type: ignore[arg-type]
                                description=sample.labels.get("description"),  # type: ignore[arg-type]
                                amount=int(sample.value),
                            )
                        ]

    return (amount_metrics, size_metrics, version_metrics)


def create_dir(path: Path) -> Path:
    """Create a directory.

    Parameters
    ----------
    path: Path
        The path for which to create a directory

    Raises
    ------
    RuntimeError
        If the path exists but is not a directory

    Returns
    -------
    Path
        The path representing the existing, absolute directory
    """
    path = path.resolve(strict=False)

    if path.exists() and not path.is_dir():
        raise RuntimeError(f"The provided path is not a directory: {path}")
    else:
        path.mkdir(parents=True, exist_ok=True)

    return path


class ProjectFiles(BaseModel):
    """A pydantic model to operate on a project's files and releases.

    Attributes
    ----------
    project_config: ProjectConfig
        A ProjectConfig instance describing the project
    upstream: Upstream
        An Upstream instance used for queries to releases of a project
    promoted_releases: List[str]
        A list of version strings
    """

    project_config: ProjectConfig
    upstream: Upstream
    promoted_releases: List[str]
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, project_config: ProjectConfig, settings: Settings) -> None:
        """Initialize an instance of ProjectFiles.

        The names of the configured maximum number of promoted releases is retrieved using an Upstream instance.
        If the project's sync_dir does not exist, it will be created.

        Parameters
        ----------
        project_config: ProjectConfig
            A ProjectConfig instance describing the project
        settings: Settings
            A Settings instance used to initialize an Upstream instance
        """
        upstream = Upstream(
            url=settings.GITLAB_URL,
            private_token=None,
            name=project_config.name,
        )
        promoted_releases = upstream.get_releases(
            max_releases=project_config.sync_config.backlog,  # type: ignore
            promoted=True,
        )
        print(f"Synchronizing release versions for {project_config.name}: " f"{', '.join(promoted_releases)}")

        create_dir(path=project_config.sync_config.directory)  # type: ignore

        super().__init__(
            project_config=project_config,
            upstream=upstream,
            promoted_releases=promoted_releases,
        )

    @classmethod
    def sync(self, project_config: ProjectConfig, settings: Settings) -> None:
        """Initialize an instance of ProjectFiles and synchronize all of its promoted_releases.

        Parameters
        ----------
        project_config: ProjectConfig
            A ProjectConfig instance describing the project
        settings: Settings
            A Settings instance used to initialize an Upstream instance
        """
        change_state: List[bool] = []

        project_files = self(
            project_config=project_config,
            settings=settings,
        )

        for child in project_files.project_config.sync_config.directory.iterdir():  # type: ignore
            if child.name.startswith(".tmp-"):
                print(f"Removing pre-existing temporary directory: {child}")
                shutil.rmtree(child)

        with tempfile.TemporaryDirectory(
            prefix=".tmp-",
            dir=project_files.project_config.sync_config.directory  # type: ignore
            if project_files.project_config.sync_config.temp_in_sync_dir  # type: ignore
            else None,
        ) as temp_dir_base_name:
            for promoted_release in project_files.promoted_releases:
                change_state += [
                    project_files._sync_version(temp_dir_base=Path(temp_dir_base_name), version=promoted_release)
                ]

        change_state += [project_files._set_latest_version_symlink()]
        change_state += [project_files._remove_obsolete_releases()]

        if any(change_state):
            project_files._set_last_update_file_timestamp()

    def _set_last_update_file_timestamp(self) -> None:
        """Write the current seconds since the epoch to a "last update file" if it is configured."""
        if self.project_config.sync_config.last_updated_file:  # type: ignore
            print(f"Updating timestamp in {self.project_config.sync_config.last_updated_file}...")  # type: ignore
            with open(self.project_config.sync_config.last_updated_file, "w") as file:  # type: ignore
                file.write(f"{int(time.time())}")

            print("Done!")

    def _remove_obsolete_releases(self) -> bool:
        """Remove obsolete releases of a project from its sync_dir."""
        state: List[bool] = []
        expected_dirs: List[Path] = []
        expected_files: List[Path] = []

        for release_type in self.project_config.releases:
            release_type_dir = self.project_config.sync_config.directory / Path(release_type.name)  # type: ignore
            print(f"Removing obsolete release files from '{release_type_dir}'...")

            expected_dirs += [release_type_dir / Path("latest")]
            expected_dirs += [
                release_type_dir / Path(f"{release_type.name}-{version}") for version in self.promoted_releases
            ]
            expected_files += [
                release_type_dir / Path(f"{release_type.name}-{version}.json") for version in self.promoted_releases
            ]
            expected_files += [
                release_type_dir / Path(f"{release_type.name}-{version}.torrent") for version in self.promoted_releases
            ]

            for file in release_type_dir.iterdir():
                if file.is_dir() and file not in expected_dirs:
                    print(f"Removing directory '{file}'")
                    shutil.rmtree(path=file)
                    state += [True]
                if file.is_file() and file not in expected_files:
                    print(f"Removing file '{file}'")
                    file.unlink()
                    state += [True]

            print("Done!")

        return any(state)

    def _set_latest_version_symlink(self) -> bool:
        """Set the symlink to the latest version in a project's sync_dir."""
        state: List[bool] = []

        if self.promoted_releases:
            latest_version = sorted(self.promoted_releases)[-1]

            for release_type in self.project_config.releases:
                latest_link = (
                    self.project_config.sync_config.directory / Path(release_type.name) / Path("latest")  # type: ignore
                )
                release_dir = Path(f"{release_type.name}-{latest_version}")
                print(f"Establishing '{latest_version}' as latest release version for '{release_type.name}'...")

                if latest_link.exists():
                    if (latest_link.is_symlink() and latest_link.readlink() != release_dir) or latest_link.is_file():
                        latest_link.unlink()
                    if not latest_link.is_symlink() and latest_link.is_dir():
                        shutil.rmtree(path=latest_link)
                if not latest_link.exists():
                    latest_link.symlink_to(release_dir)
                    state += [True]

            print("Done!")

        return any(state)

    def _sync_version(self, temp_dir_base: Optional[Path], version: str) -> bool:
        """Synchronize a project's (release) version.

        Download a project release's promotion artifact and use it to establish whether the release version has been
        synchronized fully to the project's sync_dir.
        Download the project release's build artifact, if the release version is not yet (fully) synchronized and move
        the combined artifacts to the project's sync_dir.

        Parameters
        ----------
        temp_dir_base: Optional[Path]
            The directory to use as base for creating temporary directories in when downloading and moving files
        version: str
            The project's release version

        Returns
        -------
        bool
            True if the synchronization introduced changes in the synchronization directory, False otherwise
        """
        with tempfile.TemporaryDirectory(prefix=TEMP_DIR_PREFIX, dir=temp_dir_base) as promotion_temp_dir_name:
            promotion_temp_dir = Path(promotion_temp_dir_name)
            extract_zip_file_to_parent_dir(
                path=self.upstream.download_promotion_artifact(
                    tag_name=version,
                    temp_dir=promotion_temp_dir,
                )
            )
            if self._project_version_requires_sync(version=version):
                with tempfile.TemporaryDirectory(prefix=TEMP_DIR_PREFIX, dir=temp_dir_base) as build_temp_dir_name:
                    build_temp_dir = Path(build_temp_dir_name)
                    build_artifact = self.upstream.download_release(
                        tag_name=version,
                        temp_dir=build_temp_dir,
                        job_name=self.project_config.job_name,
                    )
                    extract_zip_file_to_parent_dir(path=build_artifact)

                    for release_type in [
                        load_release_from_json_payload(path=json_payload)
                        for json_payload in promotion_temp_dir.glob("*/*.json")
                    ]:
                        self.copy_release_type_promotion_artifacts_to_build_dir(
                            release_type=release_type,
                            source_base=promotion_temp_dir,
                            destination_base=build_temp_dir / self.project_config.output_dir,
                        )

                        self.validate_release_type_files(
                            release_type=release_type,
                            path=build_temp_dir / self.project_config.output_dir,
                        )
                        self.move_release_type_to_sync_dir(
                            release=release_type,
                            source_base=build_temp_dir / self.project_config.output_dir,
                            sync_dir=self.project_config.sync_config.directory,  # type: ignore
                        )

                    return True

            return False

    def _is_release_type_synced(self, name: str, version: str) -> bool:
        """Check whether a release type of a project's release version is synchronized fully already.

        Parameters
        ----------
        name: str
            The release name
        version: str
            The release version

        Returns
        -------
        bool
            True if the project's release type in the specified version is fully synchronized, False otherwise
        """
        release_base = self.project_config.sync_config.directory / Path(name)  # type: ignore
        release_json = release_base / Path(f"{name}-{version}.json")
        if not release_json.exists():
            return False

        release = load_release_from_json_payload(path=release_json)

        if release.torrent_file and not (release_base / Path(release.torrent_file)).exists():
            return False

        release_dir = release_base / Path(f"{name}-{version}")
        for file in release.files:
            if not (release_dir / Path(file)).exists():
                return False

        return True

    def _project_version_requires_sync(
        self,
        version: str,
    ) -> bool:
        """Evaluate whether a project's release requires syncing.

        Parameters
        ----------
        version: str
            The version of the project's release

        Returns
        -------
        bool
            True if any of the release types of the project are not yet fully synchronized, False otherwise
        """
        release_type_states: List[bool] = []
        for project_release_type in self.project_config.releases:
            if not self._is_release_type_synced(
                name=project_release_type.name,
                version=version,
            ):
                release_type_states += [True]
            else:
                release_type_states += [False]

        return any(release_type_states)

    @classmethod
    def copy_release_type_promotion_artifacts_to_build_dir(
        self,
        release_type: Release,
        source_base: Path,
        destination_base: Path,
    ) -> None:
        """Copy promotion artifacts of a release type to its respective build artifact directory.

        Parameters
        ----------
        release_type: Release
            The Release instance describing the files to copy
        source_base: Path
            The base directory in which the promotion artifacts of the release type are located
        destination_base: Path
            The base directory in which the build artifacts of the release type are located
        """
        copy_signatures(
            source=source_base / Path(f"{release_type.name}/{release_type.name}-{release_type.version}"),
            destination=destination_base / Path(f"{release_type.name}/{release_type.name}-{release_type.version}"),
        )
        # move torrent file if it exists
        if release_type.torrent_file:
            (source_base / Path(f"{release_type.name}/{release_type.torrent_file}")).rename(
                destination_base / Path(f"{release_type.name}/{release_type.torrent_file}")
            )
        # move JSON payload
        (source_base / Path(f"{release_type.name}/{release_type.name}-{release_type.version}.json")).rename(
            (destination_base / Path(f"{release_type.name}/{release_type.name}-{release_type.version}.json"))
        )

    @classmethod
    def validate_release_type_files(self, release_type: Release, path: Path) -> None:
        """Validate the files of a release type.

        Parameters
        ----------
        release_type: Release
            The Release instance describing the release type
        path: Path
            The directory in which the files for the release type are located

        Raises
        ------
        RuntimeError
            If one of the files is missing
        """
        print(f"Validating release type '{release_type.name}' version '{release_type.version}'...")

        json_payload = Path(f"{release_type.name}/{release_type.name}-{release_type.version}.json")
        if not (path / json_payload).exists():
            raise RuntimeError(f"The file '{json_payload}' does not exist.")

        torrent_file = Path(f"{release_type.name}/{release_type.name}-{release_type.version}.torrent")
        if release_type.torrent_file and not (path / torrent_file).exists():
            raise RuntimeError(f"The file '{torrent_file}' does not exist.")

        for file in release_type.files:
            file_path = Path(f"{release_type.name}/{release_type.name}-{release_type.version}/{file}")
            if not (path / file_path).exists():
                raise RuntimeError(f"The file '{file_path}' does not exist.")

        print("Done!")

    @classmethod
    def move_release_type_to_sync_dir(self, release: Release, source_base: Path, sync_dir: Path) -> None:
        """Move a validated version of a release type to a sync directory.

        Parameters
        ----------
        release_type: Release
            The Release instance describing the release type
        source_base: Path
            The directory in which the files for the release type are located
        sync_dir: Path
            The directory to which the release type files are moved
        """
        print(f"Moving release type '{release.name}' version '{release.version}' to '{sync_dir}'...")

        release_type_base = sync_dir / Path(f"{release.name}")
        create_dir(release_type_base)
        destination_release_dir = release_type_base / Path(f"{release.name}-{release.version}")

        # if the destination exists already, remove it first
        if destination_release_dir.exists():
            if destination_release_dir.is_dir():
                shutil.rmtree(path=destination_release_dir)
            else:
                destination_release_dir.unlink()

        destination_release_dir.mkdir(parents=True)

        for file in release.files:
            shutil.move(
                src=source_base / Path(f"{release.name}/{release.name}-{release.version}/{file}"),
                dst=destination_release_dir / Path(f"{file}"),
            )

        if release.torrent_file:
            shutil.move(
                src=source_base / Path(f"{release.name}/{release.name}-{release.version}.torrent"),
                dst=release_type_base / Path(f"{release.name}-{release.version}.torrent"),
            )

        shutil.move(
            src=source_base / Path(f"{release.name}/{release.name}-{release.version}.json"),
            dst=release_type_base / Path(f"{release.name}-{release.version}.json"),
        )

        print("Done!")



"""Argparse module for arch-release-promotion."""

import argparse
from importlib import metadata
from sys import exit


class ArgParseFactory:
    """A factory class to create different types of argparse.ArgumentParser instances.

    Attributes
    ----------
    parser: argparse.ArgumentParser
        The instance's ArgumentParser instance, which is created with a default verbose argument

    """

    def __init__(self, prog: str = "program", description: str = "default") -> None:
        """Initialize a new ArgParseFactory."""
        self.parser = argparse.ArgumentParser(prog=prog, description=description)
        self.parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="verbose output",
        )
        self.parser.add_argument(
            "-V",
            "--version",
            action="store_true",
            help="version information",
        )

    @classmethod
    def promote(self) -> argparse.ArgumentParser:
        """Create an ArgumentParser for promotion.

        Returns
        -------
        argparse.ArgumentParser
            An ArgumentParser instance specific for promotion
        """
        instance = self(
            prog="arch-release-promotion",
            description="Download release artifacts from a project and promote them",
        )
        instance.parser.add_argument(
            "-p",
            "--project",
            type=self.non_zero_string,
            help=(
                "the project on a remote to sign (e.g. 'group/project'). "
                f"By default {instance.parser.prog} attempts to promote releases for "
                "all projects specified in its config"
            ),
        )
        instance.parser.add_argument(
            "-r",
            "--release",
            type=self.non_zero_string,
            help=(
                "the release of a project to sign (e.g. '0.1.0'). "
                f"By default {instance.parser.prog} requires user input to select a release."
            ),
        )
        if instance.parser.parse_args().version:
            print(f"{instance.parser.prog} {metadata.version('arch_release_promotion')}")
            exit(0)

        return instance.parser

    @classmethod
    def synchronize(self) -> argparse.ArgumentParser:
        """Create an ArgumentParser for synchronization.

        Returns
        -------
        argparse.ArgumentParser
            An ArgumentParser instance specific for synchronization
        """
        instance = self(
            prog="arch-release-sync",
            description="Synchronize promoted releases of a project with a local directory",
        )
        instance.parser.add_argument(
            "-p",
            "--project",
            type=self.non_zero_string,
            help=(
                "the project to synchronize from a remote (e.g. 'group/project'). "
                f"By default {instance.parser.prog} attempts to synchronize releases for "
                "all projects specified in its config"
            ),
        )

        if instance.parser.parse_args().version:
            print(f"{instance.parser.prog} {metadata.version('arch_release_promotion')}")
            exit(0)

        return instance.parser

    @classmethod
    def non_zero_string(self, input_: str) -> str:
        """Validate a string to be of non_zero length.

        Parameters
        ----------
        input_: str
            A string

        Raises
        ------
        argparse.ArgumentTypeError:
            If a string is of zero length

        Returns
        -------
        str
            A string that is of non-zero length
        """
        if len(input_) < 1:
            raise argparse.ArgumentTypeError("the provided string can not be empty")
        return input_



"""Signature handling."""

from pathlib import Path
from subprocess import run
from typing import List


def sign_files_in_dir(path: Path, developer: str, gpgkey: str, file_extensions: List[str] = []) -> None:
    """Create a detached PGP signature for one or more files in a release.

    Parameters
    ----------
    path: Path
        The path to the release
    developer: str
        The developer mbox (i.e. "First Last <user@domain.tld>")
    gpgkey: str
        The PGP key to sign with
    """
    for _file in path.iterdir():
        if _file.suffix in file_extensions:
            print(f"Creating signature for {_file}...")
            return_code = 255
            while return_code != 0:
                return_code = sign_file(path=_file.resolve(), developer=developer, gpgkey=gpgkey)
            print("Done!")


def sign_file(path: Path, developer: str, gpgkey: str) -> int:
    """Sign a file and return the status code of gpg.

    Parameters
    ----------
    path: Path
        The path to a file to sign
    developer: str
        The developer mbox (i.e. "First Last <user@domain.tld>")
    gpgkey: str
        The PGP key to sign with

    Returns
    -------
    int
        The returncode of the gpg process
    """
    return run(
        [
            "gpg",
            "--batch",
            "--no-armor",
            "--no-include-key-block",
            "--sender",
            developer,
            "--default-key",
            gpgkey,
            "--detach-sign",
            str(path),
        ],
    ).returncode
